/* FCE Ultra - NES/Famicom Emulator
*
* Copyright notice for this file:
*  Copyright (C) 2002 Xodnizel
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "common.h"
#include <Mmsystem.h> //(include Windows.h)
#pragma comment(lib, "winmm.lib")

#include "input.h"
#include "miracle.h"

static pm_module* InitPM = NULL;
nes_miracle_device* miraclep = NULL;

char MiraclePianoResponse[] = { 0xF0, 0x00, 0x00, 0x42, 0x01, 0x05, 0x01, 0x0A, 0xF7 }; // Response: Miracle Piano
char MiraclePianoQuery[] = { 0xF0, 0x00, 0x00, 0x42, 0x01, 0x04, 0xF7 }; // Query are you Miracle Piano
uint8 mididatain[0x200 * 10];// = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
char TestNote[] = { 0x90, 0x30, 0x40 };
char mididataout[100];// = { 0xF0, 0x00, 0x00, 0x42, 0x01, 0x04, 0xF7 };

#if 0

static HRESULT  midival;
static HMIDIIN hMidiIn = 0;
static HMIDIOUT hMidiOut = 0;
static MIDIHDR InMsg;
static MIDIHDR OutMsg;

char MiraclePianoResponse[] = { 0xF0, 0x00, 0x00, 0x42, 0x01, 0x05, 0x01, 0x0A, 0xF7 }; // Response: Miracle Piano
char MiraclePianoQuery[] = { 0xF0, 0x00, 0x00, 0x42, 0x01, 0x04, 0xF7 }; // Query are you Miracle Piano
char mididatain[100];// = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
char TestNote[] = { 0x90, 0x54, 0x67 };
char mididataout[100];// = { 0xF0, 0x00, 0x00, 0x42, 0x01, 0x04, 0xF7 };

static bool IsMiraclePianoHW = false;

void MidiClose(void)
{
	if (hMidiIn)
	{
		midival = midiInUnprepareHeader(hMidiIn, &InMsg, sizeof(InMsg));
		midiInStop(hMidiIn);
		midiInClose(hMidiIn);
		hMidiIn = NULL;
	}
	if (hMidiOut)
	{
		midival = midiOutUnprepareHeader(hMidiOut, &OutMsg, sizeof(OutMsg));
		midiOutClose(hMidiOut);
		hMidiOut = NULL;
	}
}

static unsigned int keys[256] = { 0, }; // with repeat
//int ctr = 0;
void MidiUpdateState(void)
{
#if 0

	unsigned char tk[256];

	midival = IDirectInputDevice7_GetDeviceState(lphMidiIn, 256, tk);
	if (tk[0]) tk[0] = 0;	//adelikat: HACK.  If a keyboard key is recognized as this, the effect is that all non assigned hotkeys are run.  This prevents the key from being used, but also prevent "hotkey explosion".  Also, they essentially couldn't use it anyway since FCEUX doesn't know it is a shift key, and it can't be assigned in the hotkeys
	// HACK because DirectInput is totally wacky about recognizing the PAUSE/BREAK key
	if (GetAsyncKeyState(VK_PAUSE)) // normally this should have & 0x8000, but apparently this key is too special for that to work
		tk[0xC5] = 0x80;

	switch (midival)
	{
	case DI_OK: //memcpy(keys,tk,256);break;
		break;

		//mbg 10/8/2008
		//previously only these two cases were handled. this made dwedit's laptop volume keys freak out.
		//we're trying this instead
	default:
		//case DIERR_INPUTLOST:
		//case DIERR_NOTACQUIRED:
		memset(tk, 0, 256);
		IDirectInputDevice7_Acquire(lphMidiIn);
		break;
	}

	//process keys
	extern int soundoptions;
#define SO_OLDUP      32

	extern int soundo;
	extern int32 fps_scale;

	for (int i = 0; i < 256; i++)
		if (tk[i])
			if (keys_nr[i] < 255)
				keys_nr[i]++; // activate key, and count up for repeat
			else
				keys_nr[i] = 255 - KEY_REPEAT_REPEATING_DELAY; // oscillate for repeat
		else
			keys_nr[i] = 0; // deactivate key

	memcpy(keys, keys_nr, sizeof(keys));

	// key-down detection
	for (int i = 0; i < 256; i++)
		if (!keys_nr[i])
		{
			keys_jd[i] = 0;
			keys_jd_lock[i] = 0;
		}
		else if (keys_jd_lock[i])
		{
		}
		else if (keys_jd[i]
			/*&& (i != 0x2A && i != 0x36 && i != 0x1D && i != 0x38)*/)
		{
			if (++keys_jd[i] > KEY_JUST_DOWN_DURATION)
			{
				keys_jd[i] = 0;
				keys_jd_lock[i] = 1;
			}
		}
		else
			keys_jd[i] = 1;

		// key repeat
		for (int i = 0; i < 256; i++)
			if ((int)keys[i] >= KEY_REPEAT_INITIAL_DELAY && !(keys[i] % KEY_REPEAT_REPEATING_DELAY))
				keys[i] = 0;

		extern uint8 autoHoldOn, autoHoldReset;
		autoHoldOn = autoHoldKey && keys[autoHoldKey] != 0;
		autoHoldReset = autoHoldClearKey && keys[autoHoldClearKey] != 0;
#endif
}

unsigned int *GetMidi(void)
{
	return(keys);
}

void CALLBACK MidiInProc(HMIDIIN hMidiIn, UINT wMsg, DWORD dwInstance, DWORD dwParam1, DWORD dwParam2)
{
	LPMIDIHDR Results;
	switch (wMsg) {
	case MIM_OPEN:
		printf("wMsg=MIM_OPEN\n");
		break;
	case MIM_CLOSE:
		printf("wMsg=MIM_CLOSE\n");
		break;
	case MIM_DATA:
		printf("wMsg=MIM_DATA, dwInstance=%08x, dwParam1(msg)=%08x, dwParam2(timestamp)=%08x\n", dwInstance, dwParam1, dwParam2);
		break;
	case MIM_LONGDATA:
		printf("wMsg=MIM_LONGDATA, dwInstance=%08x, dwParam1(lpmidihdr)=%08x, dwParam2(timestamp)=%08x\n", dwInstance, dwParam1, dwParam2);
		Results = (LPMIDIHDR)dwParam1;
		printf("wMsg=MIM_LONGDATA, size = %08x\n", Results->dwBufferLength);
		for (int i = 0; i < Results->dwBufferLength; i++)
			printf(" %02x", 0xFF & Results->lpData[i]);
		printf("\n");
		if (!IsMiraclePianoHW)
			if (0 == memcmp(MiraclePianoResponse, Results->lpData, sizeof(MiraclePianoResponse)))
				IsMiraclePianoHW = true;
		break;
	case MIM_ERROR:
		printf("wMsg=MIM_ERROR\n");
		break;
	case MIM_LONGERROR:
		printf("wMsg=MIM_LONGERROR\n");
		break;
	case MIM_MOREDATA:
		printf("wMsg=MIM_MOREDATA\n");
		break;
	default:
		printf("wMsg = unknown, dwInstance=%08x, dwParam1=%08x, dwParam2=%08x\n", dwInstance, dwParam1, dwParam2);
		break;
	}
	return;
}

//Not needed?
void CALLBACK MidiOutProc(HMIDIIN hMidiOut, UINT wMsg, DWORD dwInstance, DWORD dwParam1, DWORD dwParam2)
{
	switch (wMsg) {
	case MOM_OPEN:
		printf("wMsg=MOM_OPEN\n");
		break;
	case MOM_CLOSE:
		printf("wMsg=MOM_CLOSE\n");
		break;
	case MOM_POSITIONCB:
		printf("wMsg=MOM_POSITIONCB, dwInstance=%08x, dwParam1=%08x, dwParam2=%08x\n", dwInstance, dwParam1, dwParam2);
		break;
	case MOM_DONE:
		printf("wMsg=MOM_DONE, dwInstance=%08x, dwParam1=%08x, dwParam2=%08x\n", dwInstance, dwParam1, dwParam2);
		break;
	default:
		printf("wMsg = unknown, dwInstance=%08x, dwParam1=%08x, dwParam2=%08x\n", dwInstance, dwParam1, dwParam2);
		break;
	}
	return;
}

int GetMidiDeviceList()
{
	uint32 NumInDevs = midiInGetNumDevs();
	uint32 NumOutDevs = midiOutGetNumDevs();

	if ((NumInDevs < 1) || (NumOutDevs < 1))
	{
		FCEUD_PrintError("Midi: Error finding Midi devices.");
		return 0;
	}
	MIDIINCAPS capsi;
	for (unsigned int i = 0; i < NumInDevs; ++i) {
		midiInGetDevCaps(i, &capsi, sizeof(MIDIINCAPS));
		printf("\t%d : name = %s\n", i, capsi.szPname);
	}
	MIDIOUTCAPS capso;
	for (unsigned int i = 0; i < NumOutDevs; ++i) {
		midiOutGetDevCaps(i, &capso, sizeof(MIDIOUTCAPS));
		printf("\t%d : name = %s\n", i, capso.szPname);
	}
	return 1;
}

int OpenMidiDevices(uint32 InDev, uint32 OutDev)
{
	midival = midiInOpen(&hMidiIn, InDev, (DWORD)(void*)MidiInProc, 0, CALLBACK_FUNCTION);
	if (midival != MMSYSERR_NOERROR)
	{
		FCEUD_PrintError("Midi: Error creating midi device.");
		return 0;
	}
	midival = midiOutOpen(&hMidiOut, OutDev, (DWORD)(void*)MidiOutProc, 0, CALLBACK_FUNCTION);
	if (midival != MMSYSERR_NOERROR)
	{
		FCEUD_PrintError("Midi: Error creating midi device.");
		return 0;
	}

	midival = midiInReset(hMidiIn);
	if (midival != MMSYSERR_NOERROR)
	{
		FCEUD_PrintError("Midi: Error setting resetting timestamps.");
		return 0;
	}

	midival = midiInStart(hMidiIn);
	if (midival != MMSYSERR_NOERROR)
	{
		FCEUD_PrintError("Midi: Error setting resetting timestamps.");
		return 0;
	}

	midival = midiOutReset(hMidiOut);
	if (midival != MMSYSERR_NOERROR)
	{
		FCEUD_PrintError("Midi: Error setting resetting timestamps.");
		return 0;
	}

	size_t datasizein = sizeof(mididatain);
	size_t datasize = sizeof(mididataout);
	InMsg.lpData = mididatain;
	InMsg.dwBufferLength = datasizein;
	InMsg.dwFlags = NULL;
	OutMsg.lpData = mididataout;
	OutMsg.dwBufferLength = datasize;
	OutMsg.dwFlags = NULL;
	memcpy_s(mididataout, 100, MiraclePianoQuery, sizeof(MiraclePianoQuery));

	midival = midiInPrepareHeader(hMidiIn, &InMsg, sizeof(InMsg));
	if (midival != MMSYSERR_NOERROR)
	{
		FCEUD_PrintError("Midi: Error setting resetting timestamps.");
		return 0;
	}
	midival = midiInAddBuffer(hMidiIn, &InMsg, sizeof(InMsg));
	if (midival != MMSYSERR_NOERROR)
	{
		FCEUD_PrintError("Midi: Error setting resetting timestamps.");
		return 0;
	}

	midival = midiOutPrepareHeader(hMidiOut, &OutMsg, sizeof(OutMsg));
	if (midival != MMSYSERR_NOERROR)
	{
		FCEUD_PrintError("Midi: Error setting resetting timestamps.");
		return 0;
	}
	return 1;
}

int SendMidiMessage(char* Msg, int MsgSize)
{
	midival = midiInUnprepareHeader(hMidiIn, &InMsg, sizeof(InMsg));
	midival = midiInPrepareHeader(hMidiIn, &InMsg, sizeof(InMsg));
	midival = midiInAddBuffer(hMidiIn, &InMsg, sizeof(InMsg));
	//memcpy_s(mididataout, sizeof(mididataout), Msg, MsgSize);
	//OutMsg.dwBufferLength = MsgSize;
	midival = midiOutLongMsg(hMidiOut, &OutMsg, sizeof(OutMsg));
	if (midival != MMSYSERR_NOERROR)
	{
		FCEUD_PrintError("Midi: Error setting resetting timestamps.");
		return 0;
	}
	return 1;
}

int MidiInitialize(void)
{
	if (hMidiIn)
		return(1);

	if (GetMidiDeviceList() == 0)
		return 0;

	uint32 InDev = 0;
	uint32 OutDev = 1;
	if (OpenMidiDevices(InDev, OutDev) == 0)
		return 0;

	if (SendMidiMessage(MiraclePianoQuery, sizeof(MiraclePianoQuery)) == 0)
		return 0;
	Sleep(1000);
	if (SendMidiMessage(TestNote, sizeof(TestNote)) == 0)
		return 0;

	return 1;
}

#endif

void MidiUpdateState(void)
{
	miraclep->rcv_char();
}

void MidiClose(void)
{
	if (miraclep)
	{
		miraclep->m_midiin.close();
		miraclep->m_midiout.close();
		delete miraclep;
		miraclep = NULL;
	}
	if (InitPM)
	{
		InitPM->exit();
		InitPM = NULL;
	}
}

int GetMidiDeviceList()
{
	int num_devs = Pm_CountDevices();
	const PmDeviceInfo *pmInfo;

	printf("\n");

	if (num_devs == 0)
	{
		printf("No MIDI ports were found\n");
		return 0;
	}

	printf("MIDI input ports:\n");
	for (int i = 0; i < num_devs; i++)
	{
		pmInfo = Pm_GetDeviceInfo(i);

		if (pmInfo->input)
		{
			printf("%d %s %s\n", i, pmInfo->name, (i == Pm_GetDefaultInputDeviceID()) ? "(default)" : "");
		}
	}

	printf("\nMIDI output ports:\n");
	for (int i = 0; i < num_devs; i++)
	{
		pmInfo = Pm_GetDeviceInfo(i);

		if (pmInfo->output)
		{
			printf("%d %s %s\n", i, pmInfo->name, (i == Pm_GetDefaultOutputDeviceID()) ? "(default)" : "");
		}

	}
	return 1;
}

int MidiInitialize(void)
{
	if (InitPM)
		return(1);

	InitPM = new pm_module();
	InitPM->init();
	miraclep = new nes_miracle_device();

	if (GetMidiDeviceList() == 0)
		return 0;

	uint32 InDev = 9;
	uint32 OutDev = 19;
	const PmDeviceInfo *pmInfo = Pm_GetDeviceInfo(InDev);
	if (!miraclep->m_midiin.open_input(pmInfo->name))
		return 0;
	pmInfo = Pm_GetDeviceInfo(OutDev);
	if (!miraclep->m_midiout.open_output(pmInfo->name))
		return 0;

	for (int i = 0; i < sizeof(MiraclePianoQuery); i++)
		miraclep->m_midiout.write((MiraclePianoQuery[i]));
	Sleep(10);
	printf("pmRead=");
	while (miraclep->m_midiin.poll())
	{
		int BytesRead = miraclep->m_midiin.read(mididatain);
		for (int i = 0; i < BytesRead; i++)
			printf(" %02x", mididatain[i]);
		Sleep(1);
	}
	printf("\n");
	Sleep(1000);
	for (int i = 0; i < sizeof(TestNote); i++)
		miraclep->m_midiout.write((TestNote[i]));
	
	return 1;
}
